"use strict";
var transduce = require('transduce'), undef;

module.exports = function(_r){
  var _ = _r._,
      IGNORE = _r.IGNORE,
      as = _r.as;

  _r.wrap = wrap;
  _r.unwrap = unwrap;
  _r.deref = unwrap;
  _r.iterator = iterator;
  _r.iteratee = iteratee;
  _r.empty = empty;
  _r.append = append;
  _r.conj = append;
  _r.conjoin = append;
  _r.dispatch = dispatchXf;

  // Dispatchers
  // -----------
  var dispatch = _.reduce(
        ['iterator', 'iteratee', 'empty', 'append', 'wrap', 'unwrap'],
        function(memo, item){

        var d = function(){
          var args = arguments, fns = d._fns, i = fns.length, result,
              self = as(this);
          for(; i-- ;){
            result = fns[i].apply(self, args);
            if(result !== undef){
              return result;
            }
          }
        };

        d._fns = [];

        d.register = function(fn){
          d._fns.push(fn);
        };

        memo[item] = d;
        return memo;
      }, {});

  // Wraps a value used as source for use during chained transformation. 
  //
  // Default returns value, or _r.empty() if undefined.
  //
  // Dispatch function. To support different types,
  // call _r.unwrap.register
  function wrap(value){
    /*jshint validthis:true*/
    return dispatch.wrap.call(this, value);
  }

  wrap.register = function(fn){
    return dispatch.wrap.register(fn);
  };

  wrap.register(function(value){
    if(_.isString(value)){
      value = [value];
    } else if(value === null || value === undef){
      value = empty();
    }
    return value;
  });

  // Unwraps (deref) a possibly wrapped value
  // Default unwraps values created with _r.reduced,
  // or calls value() on chained _r transformations,
  // otherwise returns parameter.
  //
  // Dispatch function. To support different types,
  // call _r.unwrap.register
  function unwrap(value){
    return dispatch.unwrap(value);
  }

  unwrap.register = function(fn){
    return dispatch.unwrap.register(fn);
  };

  unwrap.register(function(value){
    if(as(value)){
      return value.value();
    }
    return transduce.unreduced(value);
  });

  // Returns an iterator that has next function
  // and returns {value, done}. Default looks for
  // object with iterator Symbol (or '@@iterator').
  // This is available with _r.iterator.Symbol
  //
  // Dispatch function. To support different types
  // call _r.iterator.register and supply function that returns
  // an iterator after checking the input using appropriate
  // predicates. Return undefined if not supported, so other
  // dispatched functions can be checked
  function iterator(obj){
    return dispatch.iterator(obj);
  }

  iterator.register = function(fn){
    return dispatch.iterator.register(fn);
  };

  iterator.register(transduce.iterator);

  // Mostly internal function that generates a callback from the given value.
  // For use with generating callbacks for map, filter, find, etc.
  //
  // Default returns _.iteratee.
  //
  // Dispatch function. To support different types
  // call _r.iteratee.register and supply function that returns
  // a callback after checking the input using appropriate
  // predicates. Return undefined if not supported, so other
  // dispatched functions can be checked
  function iteratee(value){
    return dispatch.iteratee(value);
  }

  iteratee.register = function(fn){
    return dispatch.iteratee.register(fn);
  };

  iteratee.register(function(value){
    if(as(value)){
      return _riteratee(value);
    }
    return _.iteratee(value);
  });

  function _riteratee(value){
    return function(item){
      return value.withSource(item).value();
    };
  }

  // Returns empty object of the same type as argument.
  // Default returns [] if _.isArray or undefined, {} if _.isObject
  // and an internal sentinel to ignore otherwise
  //
  // Dispatch function. To support different types
  // call _r.empty.register and supply function that returns
  // an empty object after checking the input using appropriate
  // predicates. Return undefined if not supported, so other
  // dispatched functions can be checked
  function empty(obj){
    return obj === IGNORE ? IGNORE : dispatch.empty(obj);
  }

  empty.register = function(fn){
    return dispatch.empty.register(fn);
  };

  empty.register(function(obj){
    if(obj === undef || _.isArray(obj) || iterator(obj)){
      return []; // array if not specified or from array
    } else if(_.isObject(obj)){
      return {}; // object if from object
    }

    // ignore by default. Default append just maintains last item.
    return IGNORE;
  });

  // Appends (conjoins) the item to the collection, and returns collection
  //
  // Dispatch function. To support different types
  // call _r.append.register and supply function that append to the object
  // (first param) with the item and optional key after checking the input
  // using appropriate predicates.
  //
  // Return undefined if not supported, so other dispatched functions can be checked
  function append(obj, item){
    // valid object and item, dispatch
    return dispatch.append(obj, item);
  }

  append.register = function(fn){
    return dispatch.append.register(fn);
  };

  append.register(function(obj, item){
    if(_.isArray(obj)){
      obj.push(item);
      return obj;
    }

    // just maintain last item
    return item;
  });

  // Reducer that dispatches to empty, unwrap and append
  function Dispatch(){}
  Dispatch.prototype.init = empty;
  Dispatch.prototype.result = unwrap;
  Dispatch.prototype.step = append;

  function dispatchXf(){
    return new Dispatch();
  }
};
